/******************************************************************************
 * Copyright (c) 2012-2018 All Rights Reserved, http://www.evocortex.com      *
 *  Evocortex GmbH                                                            *
 *  Emilienstr. 1                                                             *
 *  90489 Nuremberg                                                           *
 *  Germany                                                                   *                                           *
 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <iostream>

// IR logging interface
#include "IRLogger.h"

// IR image converter
#include "ImageBuilder.h"

// Framerate counter/reducer
#include "FramerateCounter.h"

// Monitor class
#include "IRImagerMonitor.h"

// Measurement fields
#include "IRMeasurementField.h"

#ifdef _WIN32
#include "VideoDisplay.h"
#else
#define __CALLCONV
#endif

using namespace std;
using namespace evo;

FramerateCounter _fpsThermalStream;
FramerateCounter _fpsVisibleStream;
FramerateCounter _fpsRawStream;
FramerateCounter _fpsMeasurementField;
short _decimalPlaces = 1;

void onWatchdogTimeout(IRImagerMonitor* manager, void* arg)
{
  cout << "onWatchdogTimeout" << endl;
}

void onDeviceInit(IRImagerMonitor* manager, void* arg)
{
  cout << "onDeviceInit" << endl;
  IRWatchdogParams params;
  params.cb = onWatchdogTimeout;
  params.interval = 10.0;
  params.timeout  = 1000.0;
  params.arg      = NULL;
  manager->startWatchdog(params);
  
  _decimalPlaces = manager->getImager()->getTemprangeDecimal();
}

void __CALLCONV onRawFrame(unsigned char* data, int len, IRDevice* dev)
{
  // Reduce output rate
  double fps;
  if(_fpsRawStream.trigger(&fps))
  {
    cout << "onRawFrame: " << fps << " fps" << endl;
  }
}

// Function called within process call of IRImager instance with thermal image as parameter.
// Keep this function free of heavy processing load. Otherwise the frame rate will drop down significantly for the control loop.
void __CALLCONV onThermalFrame(unsigned short* thermal, unsigned int w, unsigned int h, IRFrameMetadata meta, void* arg)
{
  // Reduce output rate
  double fps;
  if(_fpsThermalStream.trigger(&fps))
  {
    cout << "onThermalFrame: " << fps << " fps" << endl;
  }

  evo::ImageBuilder iBuilder(true, _decimalPlaces);
  iBuilder.setPaletteScalingMethod(evo::eMinMax);
  iBuilder.setPalette(evo::eIron);
  iBuilder.setData(w, h, thermal);

  unsigned char* thermalImage = new unsigned char[iBuilder.getStride() * h * 3];
  iBuilder.convertTemperatureToPaletteImage(thermalImage);
  if (arg != NULL)
  {
	  VideoDisplay* videoDisplay = (VideoDisplay*)arg;
	  videoDisplay->drawCapture(0, 0, iBuilder.getStride(), h, 24, thermalImage);
  }
  delete thermalImage;
}

// Function called within process call of IRImager instance with visible image as parameter.
// Keep this function free of heavy processing load. Otherwise the frame rate will drop down significantly for the control loop.
void __CALLCONV onVisibleFrame(unsigned char* yuyv, unsigned int w, unsigned int h, IRFrameMetadata meta, void* arg)
{
  // Reduce output rate
  double fps;
  if(_fpsVisibleStream.trigger(&fps))
  {
    cout << "onVisibleFrame: " << fps << " fps" << endl;
  }
}

// Function called within process call of IRImager instance, every time the state of the shutter flag changes.
// The flag state changes either automatically or by calling the forceFlagEvent method of IRImager.
void __CALLCONV onFlagStateChange(EnumFlagState fs, void* arg)
{
  cout << "onFlageStateChange, Flag state: " << fs << endl;
}

void onDeviceRelease(IRImagerMonitor* manager, void* arg)
{
  cout << "onDeviceRelease" << endl;
}

void onDeviceDisconnect(IRImagerMonitor* manager, void* arg)
{
  cout << "onDeviceDisconnect" << endl;
}

void onDeviceReconnect(IRImagerMonitor* manager, void* arg)
{
  cout << "onDeviceReconnect" << endl;

  manager->restartWatchdog();
}

void onMeasurementField(IRMeasurementField* field)
{
  // Reduce output rate
  double fps;
  if(_fpsMeasurementField.trigger(&fps))
  {
    cout << "Measurement field #" << field->_id << ", tMin: "  << field->_tMin  <<
                                                   ", tMean: " << field->_tMean <<
                                                   ", tMax: "  << field->_tMax  << endl;
  }
}

void onProcessExit(void* arg)
{
  // This event can be used to synchronize received data.
  // It is the very last callback for a certain image.
}

int main (int argc, char* argv[])
{
  if(argc!=2)
  {
    cout << "usage: " << argv[0] << " <xml configuration file>" << endl;
    return -1;
  }

  IRLogger::setVerbosity(IRLOG_ERROR, IRLOG_OFF);

#ifdef _WIN32
  char* args = argv[1];
  // Windows SDK is compiled using 16-bit Unicode characters
  size_t argSize = strlen(args) + 1, converted;
  wchar_t* argPath = new wchar_t[argSize];
  mbstowcs_s(&converted, argPath, argSize, args, argSize);
#else
  char* argPath = argv[1];
#endif
  // Read parameters from xml file
  IRDeviceParams params;
  if(!IRDeviceParamsReader::readXML(argPath, params))
    return -1;

  IRImagerMonitor monitor;
  monitor.setRawFrameCallback(onRawFrame);
  monitor.setThermalFrameCallback(onThermalFrame);
  monitor.setVisibleFrameCallback(onVisibleFrame);
  monitor.setFlagStateCallback(onFlagStateChange);
  monitor.setProcessExitCallback(onProcessExit);
  monitor.registerEventCallback(IREventInit,       onDeviceInit);
  monitor.registerEventCallback(IREventRelease,    onDeviceRelease);
  monitor.registerEventCallback(IREventDisconnect, onDeviceDisconnect);
  monitor.registerEventCallback(IREventReconnect,  onDeviceReconnect);
  monitor.init(params, NULL);

  int id         = 0;
  int x          = 50;
  int y          = 50;
  int fieldSize  = 20;
  int emissivity = 1.0;
  IRMeasurementField* field = new IRMeasurementField(id, x, y, fieldSize, fieldSize, emissivity, 1.0, onMeasurementField, NULL);
  monitor.addMeasurementField(field);

#ifdef _WIN32
  IRImager* imager = monitor.getImager();
  VideoDisplay vd(imager->getWidth(), imager->getHeight());
  vd.showDetach();
  monitor.setFrameArg(&vd);
#endif

  IRWatchdogParams wparams;
  wparams.cb       = onWatchdogTimeout;
  wparams.interval = 10.0;
  wparams.timeout  = 1000.0;
  monitor.startWatchdog(wparams);
  monitor.run(params, &vd);

  cout << "Exiting application" << endl;

  return 1;
}
